
class OpBot(tuple):
    """OpBot Class: Encapsulate data of an Uabot Operation
    
    Inheritance from tuple:
    \t- [0]: Opcode
    \t- [1]: Number of arguments to pass
    \t- [2]: Number of values returned
    \t- [3]: Key Opcode
    \t- [4]: Description opcode"""

    def __new__(cls, opcode, narg, nret, key, desc=None):
        return tuple.__new__(cls, (opcode, narg, nret, key.upper(), desc))
    
    def __str__(self):
        return "<Opcode: 0x%02x, Args: %d, KeyOpcode: \'%s\'>" % (self[0], self[1], self[3])
 
    def getOpcode(self): 
        "getOpcode() - Get numeric opcode"
        return self[0]

    def getNumArgs(self):
        "Number of arguments required"
        return self[1]

    def getNumRet(self):
        "Number of values returned"
        return self[2]
    
    def getKeyOpcode(self): 
        "Get string key opcode"
        return self[3]

    def getDescription(self):
        "Get an info string about opcode"
        return self[4]

    def genOp(self, *args): 
        if len(args) != self[1]:
            raise Exception("Incorrect number of arguments line")
        else:
            str = chr(self[0])
            for v in args:
                str += chr(v)
            return str


class SetOpBot:
    """SetOpBot Class: Set of aviable Uabot operations
    
    Acces to operation:
    \t- By numerical opcode
    \t- By key string opcode"""
    __OPSET = ( OpBot(0, 0, 0, "RST", "Reset operation"),
                OpBot(1, 0, -1, "TSTM", "Test M-Task"),
                OpBot(2, 2, 0, "MOT", "Motor operation"),
                OpBot(3, 2, 0, "LED", "Set/Clear leds operation"),
                OpBot(4, 2, 0, "OSC", "Oscillation lighting leds operation"),
                OpBot(5, 2, 0, "SPK", "Speaker operation"),
                OpBot(6, 0, -1, "TSTS", "Test S-Task"),
                OpBot(7, 0, 0, "WCM", "Calibrate Motors operation"),
                OpBot(8, 0, 2, "IRL", "IR Lateral operation"),
                OpBot(9, 0, 2, "IRF", "IR Frontal operation"),
                OpBot(10, 0, 2, "LDR", "LDR operation"),
                OpBot(11, 0, 1, "LFW", "Line Follower operation"),
                OpBot(12, 0, 1, "MIC", "Microphone operation"),
                OpBot(13, 0, 8, "AOP", "All sensor operations"),
                OpBot(14, 0, 0, "NOP", "No operation"),
            )
    
    __OPDICT = dict([ (v.getKeyOpcode(), v) for v in __OPSET ])

    __GROUP = { "SENSOR": dict([ ("IRL", __OPDICT["IRL"]),
                                 ("IRF", __OPDICT["IRF"]),
                                 ("LDR", __OPDICT["LDR"]),
                                 ("LFW", __OPDICT["LFW"]),
                                 ("MIC", __OPDICT["MIC"]), ]),

                "MISC"  : dict([ ("LED", __OPDICT["LED"]),
                                 ("OSC", __OPDICT["OSC"]),
                                 ("SPK", __OPDICT["SPK"]), ]),

                "CTRL"  : dict([ ("RST", __OPDICT["RST"]),
                                 ("TSTM", __OPDICT["TSTM"]),
                                 ("TSTS", __OPDICT["TSTS"]),
                                 ("WCM", __OPDICT["WCM"]),
                                 ("NOP", __OPDICT["NOP"]),]) }
        
    def __getitem__(self, key):
        try:
            if type(key) is int: return SetOpBot.__OPSET[key]
            elif type(key) is str: return SetOpBot.__OPDICT[key.upper()]
            else: raise ValueError("Wrong key value, only int or str type")
        except IndexError:
            raise IndexError("Index out of bound")
        except KeyError:
            raise KeyError("Key not found")
         
    def __setitem__(self, key, value):
        raise TypeError("Forbidden item assignment")

    def __str__(self):
        s = str()
        for op in SetOpBot.__OPSET:
            s += op.__str__() + ":: " + op.getDescription() + "\n"
        return s

    def listOpcodes(self):
        return self.__OPDICT.keys() 
        
    def group(self, key):
        return SetOpBot.__GROUP[key]

    
